function benchmark( callback, iterations )
    if iterations == nil then
        iterations = 1;
    end
    local t = GameGetRealWorldTimeSinceStarted();
    for i=1,iterations do
        callback();
    end
    return GameGetRealWorldTimeSinceStarted() - t;
end

function map(func, array)
    local new_array = {};
    for i,v in ipairs(array) do
        new_array[i] = func(v);
    end
    return new_array;
end


function DoFileEnvironment( filepath, environment )
    if environment == nil then environment = {} end
    local f = loadfile( filepath );
    local set_f = setfenv( f, setmetatable( environment, { __index = _G } ) );
    local status,result = pcall( set_f );
    if status == false then print_error( "do file environment for "..filepath..": "..result ); end
    return environment;
end

function ModTextFileAppend( left, right )
    ModTextFileSetContent( left, (ModTextFileGetContent( left ) or "") .. "\r\n" .. (ModTextFileGetContent( right ) or "") );
end

function ModTextFilePrepend( left, right )
    ModTextFileSetContent( left, (ModTextFileGetContent( right ) or "") .. "\r\n" .. (ModTextFileGetContent( left ) or "") );
end

function PackString( separator, ... )
	local string = {};
	for n=1,select( '#' , ... ) do
		local j = select( n, ... );
		local text = nil;
		if type(j) == "boolean" then
			if j then
			   text = "true";
			else
				text = "false";
			end
        else
            if type(j) == "string" and #j == 0 then
                text = "\"\"";
            else
                text = tostring( j ) or "nil";
            end
        end
		string[n] = text;
	end
	return table.concat( string, separator or "" );
end

function Log( ... )
    print( PackString(" ", ... ) );
end

function LogCompact( ... )
    local length = 0;
    local log = {};
    for k,v in pairs( {...} ) do
        table.insert( log, v );
        length = length + #v;
        if length > 80 then
            table.insert(log,"\n");
            length = 0;
        end
    end
    Log( unpack( log ) );
end

local screen_log_queue = {};
local screen_log_max = 20;
local screen_log_interval = 0.333;
function LogScreen( ... )
    table.insert( screen_log_queue, PackString( " ", ... ) );
end

function RenderLog( gui )
    local start_index = math.floor(GameGetRealWorldTimeSinceStarted()/screen_log_interval);
    for i=1,screen_log_max do
        local adjusted_index = (i + start_index) % #screen_log_queue;
        if screen_log_queue[adjusted_index] ~= nil then
            GuiText( gui, 0,0, tostring(screen_log_queue[adjusted_index]) );
        end
        GuiLayoutAddVerticalSpacing( gui );
    end
    screen_log_queue = {};
end

function LogTable( t )
    if type(t) == "table" then
        for k,v in pairs ( t ) do
            if type(v) == "table" then
                LogTable( v );
            else
                Log( k,v );
            end
        end
    else
        Log( t );
    end
end

function table.dump(o)
    if type(o) == 'table' then
        local s = '{ '..string.char(10)
        for k,v in pairs(o) do
           if type(k) ~= 'number' then k = '"'..k..'"' end
           s = s .. '['..k..'] = ' .. table.dump(v) .. ',' ..string.char(10)
        end
        return s .. '} '..string.char(10)
     else
        if(type(o) == "string")then
            return '"'..o..'"'
        else
            return tostring(o)
        end
     end
end

function LogTableCompact( t, show_keys )
    if type(t) == "table" then
        local length = 0;
        local log = {};
        for k,v in pairs( t ) do
            local join = PackString( "=", k, v );
            table.insert( log, join );
            length = length + #join;
            if length > 80 then
                table.insert(log,"\n");
                length = 0;
            end
        end
        Log( unpack( log ) );
    else
        Log( t );
    end
end

function ExploreGlobals()
    local g = {};
    for k,v in pairs(_G) do
        if type(v) ~= "function" then
            g[k] = v;
        end
    end
    LogTable(g);
end

function EntityGetNamedChild( entity_id, name )
    local children = EntityGetAllChildren( entity_id ) or {};
	if children ~= nil then
		for index,child_entity in pairs( children ) do
			local child_entity_name = EntityGetName( child_entity );
			
			if child_entity_name == name then
				return child_entity;
            end
        end
    end
end

function EntityGetChildrenWithTag( entity_id, tag )
    local valid_children = {};
    local children = EntityGetAllChildren( entity_id ) or {};
    for index, child in pairs( children ) do
        if EntityHasTag( child, tag ) then
            table.insert( valid_children, child );
        end
    end
    return valid_children;
end

function ComponentGetValueDefault( component_id, key, default )
    local value = ComponentGetValue2( component_id, key );
    if value ~= nil then
        return value;
    end
    return default;
end

function FindEntityInInventory( inventory, entity )
    local inventory_items = EntityGetAllChildren( inventory )  or {};
		
    -- remove default items
    if inventory_items ~= nil then
        for i,item_entity in ipairs( inventory_items ) do
            --Log( i, item_entity );
        end
        --    GameKillInventoryItem( player_entity, item_entity )
        --end
    end
end

function TryGivePerk( player_entity_id, ... )
    for index,perk_id in pairs( {...} ) do
        local perk_entity = perk_spawn( x, y, perk_id );
        if perk_entity ~= nil then
            perk_pickup( perk_entity, player_entity_id, EntityGetName( perk_entity ), false, false );
        end
    end
end

function TryAdjustDamageMultipliers( entity, resistances )
    local damage_models = EntityGetComponent( entity, "DamageModelComponent" );
    if damage_models ~= nil then
        for index,damage_model in pairs( damage_models ) do
            for damage_type,multiplier in pairs( resistances ) do
                local resistance = ComponentObjectGetValue2( damage_model, "damage_multipliers", damage_type );
                resistance = resistance * multiplier;
                ComponentObjectSetValue2( damage_model, "damage_multipliers", damage_type, resistance );
            end
        end
    end
end


--[[
    ice
    electricity
    radioactive
    slice
    projectile
    healing
    physics_hit
    explosion
    poison
    melee
    drill
    fire
]]

function TryAdjustMaxHealth( entity, callback )
    local damage_models = EntityGetComponent( entity, "DamageModelComponent" );
    if damage_models ~= nil then
        for index,damage_model in pairs( damage_models ) do
            local current_hp = ComponentGetValue2( damage_model, "hp" );
            local max_hp = ComponentGetValue2( damage_model, "max_hp" );
            local new_max = callback( max_hp, current_hp );
            local regained = new_max - current_hp;
            ComponentSetValue2( damage_model, "max_hp", new_max );
            ComponentSetValue2( damage_model, "hp", current_hp + regained );
        end
    end
end

function GetInventoryQuickActiveItem( entity )
    if entity ~= nil then
        local component = EntityGetFirstComponent( entity, "Inventory2Component" );
        if component ~= nil then
            return ComponentGetValue2( component, "mActiveItem" );
        end
    end
end

function EntityIterateComponentsByType( entity, component_type_name, callback )
    local matched_components = {};
    local components = EntityGetAllComponents( entity ) or {};
    for _,component in pairs(components) do
        if ComponentGetTypeName( component ) == component_type_name then
            table.insert( matched_components, component );
        end
    end
    if callback ~= nil then
        for _,component in pairs(matched_components) do
            callback( component );
        end
    end
    return matched_components;
end

function EntityGetHitboxCenter( entity )
    local tx, ty = EntityGetTransform( entity );
    local hitbox = EntityGetFirstComponent( entity, "HitboxComponent" );
    if hitbox ~= nil then
        local width = ComponentGetValue2( hitbox, "aabb_max_x" ) - ComponentGetValue2( hitbox, "aabb_min_x" );
        local height = ComponentGetValue2( hitbox, "aabb_max_y" ) - ComponentGetValue2( hitbox, "aabb_min_y" );
        tx = tx + ComponentGetValue2( hitbox, "aabb_min_x" ) + width * 0.5;
        ty = ty + ComponentGetValue2( hitbox, "aabb_min_y" ) + height * 0.5;
    end
    return tx, ty;
end

function EntityGetFirstHitboxSize( entity, fallbackWidth, fallbackHeight )
    local hitbox = EntityGetFirstComponent( entity, "HitboxComponent" );

    local width = fallbackWidth or 0;
    local height = fallbackHeight or 0;
    if hitbox ~= nil then
        width =  ComponentGetValue2( hitbox, "aabb_max_x" ) - ComponentGetValue2( hitbox, "aabb_min_x" );
        height =  ComponentGetValue2( hitbox, "aabb_max_y" ) - ComponentGetValue2( hitbox, "aabb_min_y" );
    end
    return width, height;
end

function EntitySetVelocity( entity, x, y )
    EntityIterateComponentsByType( entity, "VelocityComponent", function( component )
        ComponentSetValue2( component, "mVelocity", x, y );
    end );
end


function EaseAngle( angle, target_angle, easing )
    local dir = (angle - target_angle) / (math.pi*2);
    dir = dir - math.floor(dir + 0.5);
    dir = dir * (math.pi*2);
    return angle - dir * easing;
end

function WeightedRandom( items, weights, sum )
    if weights == nil  then
        weights = {};
        for k,v in pairs( items ) do 
            weights[k] = 1;
        end
        sum = items.length;
    end
    if sum == nil then
        sum = 0;
        for k,v in pairs( weights ) do
            sum = sum + v;
        end
    end
    if sum <= 0 then
        return nil;
    end
    local random = Random() * sum;
    for k,v in pairs(items) do
        local weight = weights[k];
        if random <= weight then
            return items[k];
        end
        random = random - weight;
    end
end

function WeightedRandomTable( entries )
    local sum = 0;
    for k,v in pairs( entries ) do
        sum = sum + v;
    end
    if sum <= 0 then
        return nil;
    end
    local random = Random() * sum;
    for k,v in pairs( entries ) do
        if random <= v then
            return k;
        end
        random = random - v;
    end
end

function EntityHasScript( entity, filepath, script_type )
    if script_type == nil then
        script_type = "script_source_file";
    end
    for _,lua_component in pairs( EntityGetComponentIncludingDisabled(entity,"LuaComponent") or {}) do
        if ComponentGetValue2( lua_component, script_type ) == filepath then
            return true;
        end
    end
    return false;
end

function EntityAddScript( entity, filepath, script_type, component_settings )
    if not EntityHasScript( entity, filepath, script_type ) then
        component_settings = component_settings or {};
        component_settings[script_type] = filepath;
        EntityAddComponent2( entity, "LuaComponent", component_settings );
        return true;
    end
    return false;
end

local _ModSettingGet = ModSettingGet;
function ModSettingGet( setting, default )
    local mod_setting_value = _ModSettingGet( setting );
    if mod_setting_value ~= nil then
        return mod_setting_value;
    else
        return default;
    end
end